from __future__ import annotations

import builtins
import datetime
import enum
import functools
import json
import os
import typing
import warnings

import fsspec
import pandas as pd
import polars as pl
import pydantic
import tlz
from pydantic import ConfigDict
from typing_extensions import Self

from ._search import search, search_apply_require_all_on

__framereaders__ = [pl, pd]
__filetypes__ = ['csv', 'csv.bz2', 'csv.gz', 'csv.zip', 'csv.xz', 'parquet']


def _allnan_or_nonan(df, column: str) -> bool:
    """Check if all values in a column are NaN or not NaN

    Returns
    -------
    bool
        Whether the dataframe column has all NaNs or no NaN valles

    Raises
    ------
    ValueError
        When the column has a mix of NaNs non NaN values
    """
    if df[column].isnull().all():
        return False
    if df[column].isnull().any():
        raise ValueError(
            f'The data in the {column} column should either be all NaN or there should be no NaNs'
        )
    return True


class AggregationType(str, enum.Enum):
    join_new = 'join_new'
    join_existing = 'join_existing'
    union = 'union'

    model_config = ConfigDict(validate_assignment=True)


class DataFormat(str, enum.Enum):
    netcdf = 'netcdf'
    zarr = 'zarr'
    zarr2 = 'zarr2'
    zarr3 = 'zarr3'
    reference = 'reference'
    opendap = 'opendap'

    model_config = ConfigDict(validate_assignment=True)


class Attribute(pydantic.BaseModel):
    column_name: pydantic.StrictStr
    vocabulary: pydantic.StrictStr = ''

    model_config = ConfigDict(validate_assignment=True)


class Assets(pydantic.BaseModel):
    column_name: pydantic.StrictStr
    format: DataFormat | None = None
    format_column_name: pydantic.StrictStr | None = None

    model_config = ConfigDict(validate_assignment=True)

    @pydantic.model_validator(mode='after')
    def _validate_data_format(cls, model):
        data_format, format_column_name = model.format, model.format_column_name
        if data_format is not None and format_column_name is not None:
            raise ValueError('Cannot set both format and format_column_name')
        elif data_format is None and format_column_name is None:
            raise ValueError('Must set one of format or format_column_name')
        return model


class Aggregation(pydantic.BaseModel):
    type: AggregationType
    attribute_name: pydantic.StrictStr
    options: dict = {}

    model_config = ConfigDict(validate_assignment=True)


class AggregationControl(pydantic.BaseModel):
    variable_column_name: pydantic.StrictStr
    groupby_attrs: list[pydantic.StrictStr]
    aggregations: list[Aggregation] = []

    model_config = ConfigDict(validate_default=True, validate_assignment=True)


class ESMCatalogModel(pydantic.BaseModel):
    """
    Pydantic model for the ESM data catalog defined in https://git.io/JBWoW
    """

    esmcat_version: pydantic.StrictStr
    attributes: list[Attribute]
    assets: Assets
    aggregation_control: AggregationControl | None = None
    id: str = ''
    catalog_dict: list[dict] | None = None
    catalog_file: pydantic.StrictStr | None = None
    description: pydantic.StrictStr | None = None
    title: pydantic.StrictStr | None = None
    last_updated: datetime.datetime | datetime.date | None = None
    _df: pd.DataFrame | None = pydantic.PrivateAttr()
    _frames: FramesModel | None = pydantic.PrivateAttr()
    _iterable_dtype_map: dict[str, str] = pydantic.PrivateAttr(default_factory=dict)

    model_config = ConfigDict(arbitrary_types_allowed=True, validate_assignment=True)

    @pydantic.model_validator(mode='after')
    def validate_catalog(cls, model):
        catalog_dict, catalog_file = model.catalog_dict, model.catalog_file
        if catalog_dict is not None and catalog_file is not None:
            raise ValueError('catalog_dict and catalog_file cannot be set at the same time')

        return model

    def __setattr__(self, name, value):
        """If we manually set _df, we need to propagate the change to _frames"""
        if name == '_df':
            self._frames = FramesModel(df=value)
        return super().__setattr__(name, value)

    @classmethod
    def from_dict(cls, data: dict) -> ESMCatalogModel:
        esmcat = data['esmcat']
        df = data['df']
        if 'last_updated' not in esmcat:
            esmcat['last_updated'] = None
        cat = cls.model_validate(esmcat)
        cat._df = df
        cat._frames = FramesModel(df=df)
        return cat

    def save(
        self,
        name: str,
        *,
        directory: str | None = None,
        catalog_type: str = 'dict',
        file_format: str = 'csv',
        write_kwargs: dict | None = None,
        to_csv_kwargs: dict | None = None,
        json_dump_kwargs: dict | None = None,
        storage_options: dict[str, typing.Any] | None = None,
    ) -> None:
        """
        Save the catalog to a file.

        Parameters
        -----------
        name: str
            The name of the file to save the catalog to.
        directory: str
            The directory or cloud storage bucket to save the catalog to.
            If None, use the current directory.
        catalog_type: str
            The type of catalog to save. Whether to save the catalog table as a dictionary
            in the JSON file or as a separate CSV file. Valid options are 'dict' and 'file'.
        file_format: str
            The file format to use when saving the catalog table. Either 'csv' or 'parquet'.
            If catalog_type is 'dict', this parameter is ignored.
        to_csv_kwargs : dict, optional
            Additional keyword arguments passed through to the :py:meth:`~pandas.DataFrame.to_csv` method.
            Compression is currently ignored when serializing to parquet.
        json_dump_kwargs : dict, optional
            Additional keyword arguments passed through to the :py:func:`~json.dump` function.
        storage_options: dict
            fsspec parameters passed to the backend file-system such as Google Cloud Storage,
            Amazon Web Service S3.

        Notes
        -----
        Large catalogs can result in large JSON files. To keep the JSON file size manageable, call with
        `catalog_type='file'` to save catalog as a separate CSV file.

        """

        if to_csv_kwargs is not None:
            warnings.warn(
                'to_csv_kwargs is deprecated and will be removed in a future version. '
                'Please use write_kwargs instead.',
                DeprecationWarning,
                stacklevel=2,
            )
            if write_kwargs is not None:
                raise ValueError(
                    'Cannot provide both `read_csv_kwargs` and `write_kwargs`. '
                    'Please use `write_kwargs`.'
                )
            write_kwargs = to_csv_kwargs

        write_kwargs = write_kwargs or {}

        if catalog_type not in {'file', 'dict'}:
            raise ValueError(
                f'catalog_type must be either "dict" or "file". Received catalog_type={catalog_type}'
            )

        # Check if the directory is None, and if it is, set it to the current directory
        if directory is None:
            directory = os.getcwd()

        # Configure the fsspec mapper and associated filenames
        storage_options = storage_options if storage_options is not None else {}
        mapper = fsspec.get_mapper(f'{directory}', **storage_options)
        fs = mapper.fs
        csv_file_name = fs.unstrip_protocol(f'{mapper.root}/{name}.csv')
        json_file_name = fs.unstrip_protocol(f'{mapper.root}/{name}.json')

        data = self.model_dump().copy()
        for key in {'catalog_dict', 'catalog_file'}:
            data.pop(key, None)
        data['id'] = name
        data['last_updated'] = datetime.datetime.now(datetime.timezone.utc).strftime(
            '%Y-%m-%dT%H:%M:%SZ'
        )

        _tmp_df = self.df.copy(deep=True)

        for colname, dtype in self._iterable_dtype_map.items():
            _tmp_df[colname] = _tmp_df[colname].apply(getattr(builtins, dtype))

        if catalog_type == 'file':
            csv_kwargs: dict[str, typing.Any] = {'index': False}
            csv_kwargs |= write_kwargs or {}
            compression = csv_kwargs.get('compression', '')
            extensions = {'gzip': '.gz', 'bz2': '.bz2', 'zip': '.zip', 'xz': '.xz'}
            if file_format == 'csv':
                csv_file_name = f'{csv_file_name}{extensions.get(compression, "")}'
                data['catalog_file'] = str(csv_file_name)
                with fs.open(csv_file_name, 'wb') as csv_outfile:
                    _tmp_df.to_csv(csv_outfile, **csv_kwargs)
            elif file_format == 'parquet':
                pq_file_name = f'{csv_file_name.rstrip(".csv")}.parquet'
                data['catalog_file'] = str(pq_file_name)
                write_kwargs.pop('compression', None)
                with fs.open(pq_file_name, 'wb') as pq_outfile:
                    _tmp_df.to_parquet(pq_outfile, **write_kwargs)
        else:
            data['catalog_dict'] = _tmp_df.to_dict(orient='records')

        with fs.open(json_file_name, 'w') as outfile:
            json_kwargs = {'indent': 2}
            json_kwargs |= json_dump_kwargs or {}
            json.dump(data, outfile, **json_kwargs)  # type: ignore[arg-type]

        print(f'Successfully wrote ESM catalog json file to: {json_file_name}')

    @classmethod
    def load(
        cls,
        json_file: str | pydantic.FilePath | pydantic.AnyUrl,
        storage_options: dict[str, typing.Any] | None = None,
        read_kwargs: dict[str, typing.Any] | None = None,
    ) -> ESMCatalogModel:
        """
        Loads the catalog from a file

        Parameters
        -----------
        json_file: str or pathlib.Path
            The path to the json file containing the catalog
        storage_options: dict
            fsspec parameters passed to the backend file-system such as Google Cloud Storage,
            Amazon Web Service S3.
        read_kwargs : dict, optional
            Additional keyword arguments passed through to the :py:func:`~pandas.read_csv` function, if the
            datastore is saved in csv format, or :py:func:`~pandas.read_parquet` if the datastore is saved in
            parquet format.

        """
        storage_options = storage_options if storage_options is not None else {}
        read_kwargs = read_kwargs or {}
        json_file = str(json_file)  # We accept Path, but fsspec doesn't.
        _mapper = fsspec.get_mapper(json_file, **storage_options)

        with fsspec.open(json_file, **storage_options) as fobj:
            data = json.loads(fobj.read())
            if 'last_updated' not in data:
                data['last_updated'] = None
            cat = cls.model_validate(data)
            if cat.catalog_file:
                cat._frames = cat._df_from_file(cat, _mapper, storage_options, read_kwargs)
            else:
                cat._frames = FramesModel(
                    lf=pl.LazyFrame(cat.catalog_dict),
                    pl_df=pl.DataFrame(cat.catalog_dict),
                    df=pl.DataFrame(cat.catalog_dict).to_pandas(),
                )

            return cat

    def _df_from_file(
        self,
        cat: ESMCatalogModel,
        _mapper: fsspec.FSMap,
        storage_options: dict[str, typing.Any],
        read_kwargs: dict[str, typing.Any],
    ) -> FramesModel:
        """
        Read the catalog file from disk, falling back to pandas for bz2 files which
        polars can't read.

        Returns a FramesModel, which contains at least one of:
        - a polars LazyFrame
        - a polars DataFrame
        - a pandas DataFrame

        , as well as handling dataframe related methods, eg. columns_with_iterables.

        Parameters
        ----------
        cat: ESMCatalogModel
            The catalog model
        _mapper: fsspec mapper
            A fsspec mapper object
        storage_options: dict
            fsspec parameters passed to the backend file-system such as Google Cloud Storage,
            Amazon Web Service S3.
        read_kwargs: dict
            Additional keyword arguments passed through to the :py:func:`~pandas.read_csv` function.

        Returns
        -------
        FramesModel:
            A pydantic model containing at least one of a pandas/polars dataframe
            and a polars lazyframe
        """
        if _mapper.fs.exists(cat.catalog_file):
            csv_path = cat.catalog_file
        else:
            csv_path = f'{os.path.dirname(_mapper.root)}/{cat.catalog_file}'
        cat.catalog_file = csv_path

        reader = CatalogFileDataReader(cat.catalog_file, storage_options, **read_kwargs)
        self._iterable_dtype_map = reader.dtype_map
        return reader.frames

    @property
    def lf(self) -> pl.LazyFrame:
        """Return a `pl.LazyFrame` containing the catalog, creating it if necessary"""
        return self._frames.lazy  # type: ignore[union-attr]

    @property
    def pl_df(self) -> pl.DataFrame:
        """Return a `pl.DataFrame` containing the catalog, creating it if necessary"""
        return self._frames.polars  # type: ignore[union-attr]

    @property
    def df(self) -> pd.DataFrame:
        """Return the `pd.DataFrame` containing the catalog, creating it if necessary"""
        return self._frames.pandas  # type: ignore[union-attr]

    @property
    def columns_with_iterables(self) -> set[str]:
        """Return a set of columns that have iterables."""
        return self._frames.columns_with_iterables  # type: ignore[union-attr]

    @property
    def has_multiple_variable_assets(self) -> bool:
        """Return True if the catalog has multiple variable assets."""
        if self.aggregation_control:
            return self.aggregation_control.variable_column_name in self.columns_with_iterables
        return False

    @property
    def grouped(self) -> pd.core.groupby.DataFrameGroupBy | pd.DataFrame:
        if self.aggregation_control:
            if self.aggregation_control.groupby_attrs:
                self.aggregation_control.groupby_attrs = list(
                    filter(
                        functools.partial(_allnan_or_nonan, self.df),
                        self.aggregation_control.groupby_attrs,
                    )
                )

            if self.aggregation_control.groupby_attrs and set(
                self.aggregation_control.groupby_attrs
            ) != set(self.df.columns):
                return self.df.groupby(self.aggregation_control.groupby_attrs)
        cols = list(
            filter(
                functools.partial(_allnan_or_nonan, self.df),
                self.df.columns,
            )
        )
        return self.df.groupby(cols)

    def _construct_group_keys(self, sep: str = '.') -> dict[str, str | tuple[str]]:
        internal_keys = self.grouped.groups.keys()
        public_keys = map(
            lambda key: key if isinstance(key, str) else sep.join(str(value) for value in key),
            internal_keys,
        )

        return dict(zip(public_keys, internal_keys))

    def _unique(self) -> dict:
        def _find_unique(series):
            values = series.dropna()
            if series.name in self.columns_with_iterables:
                values = tlz.concat(values)
            return list(tlz.unique(values))

        data = self.df[self.df.columns]
        if data.empty:
            return {col: [] for col in self.df.columns}
        else:
            return data.apply(_find_unique, result_type='reduce').to_dict()

    def unique(self) -> pd.Series:
        """Return a series of unique values for each column in the catalog."""
        return pd.Series(self._unique())

    def nunique(self) -> pd.Series:
        """Return a series of the number of unique values for each column in the catalog."""

        return self._frames.nunique()  # type: ignore[union-attr]

    def search(
        self,
        *,
        query: QueryModel | dict[str, typing.Any],
        require_all_on: str | list[str] | None = None,
    ) -> pd.DataFrame:
        """
        Search for entries in the catalog.

        Parameters
        ----------
        query: dict, optional
            A dictionary of query parameters to execute against the dataframe.
        require_all_on : list, str, optional
            A dataframe column or a list of dataframe columns across
            which all entries must satisfy the query criteria.
            If None, return entries that fulfill any of the criteria specified
            in the query, by default None.

        Returns
        -------
        catalog: ESMCatalogModel
            A new catalog with the entries satisfying the query criteria.

        """

        _query = (
            query
            if isinstance(query, QueryModel)
            else QueryModel(
                query=query, require_all_on=require_all_on, columns=self.df.columns.tolist()
            )
        )

        results = search(
            df=self.df, query=_query.query, columns_with_iterables=self.columns_with_iterables
        )
        if _query.require_all_on is not None and not results.empty:
            results = search_apply_require_all_on(
                df=results,
                query=_query.query,
                require_all_on=_query.require_all_on,
                columns_with_iterables=self.columns_with_iterables,
            )
        return results


class QueryModel(pydantic.BaseModel):
    """A Pydantic model to represent a query to be executed against a catalog."""

    query: dict[pydantic.StrictStr, typing.Any | list[typing.Any]]
    columns: list[str]
    require_all_on: str | list[typing.Any] | None = None

    # TODO: Seem to be unable to modify fields in model_validator with
    # validate_assignment=True since it leads to recursion
    model_config = ConfigDict(validate_assignment=False)

    @pydantic.model_validator(mode='after')
    def validate_query(cls, model):
        query = model.query
        columns = model.columns
        require_all_on = model.require_all_on

        if query:
            for key in query:
                if key not in columns:
                    raise ValueError(f'Column {key} not in columns {columns}')
        if isinstance(require_all_on, str):
            model.require_all_on = [require_all_on]
        if require_all_on is not None:
            for key in model.require_all_on:
                if key not in columns:
                    raise ValueError(f'Column {key} not in columns {columns}')
        _query = query.copy()
        for key, value in _query.items():
            if isinstance(value, str | int | float | bool) or value is None or value is pd.NA:
                _query[key] = [value]

        model.query = _query
        return model


class FramesModel(pydantic.BaseModel):
    """A Pydantic model to represent our collection of dataframes - pandas, polars,
    and lazyframe."""

    df: pd.DataFrame | None = None
    pl_df: pl.DataFrame | None = None
    lf: pl.LazyFrame | None = None

    model_config = ConfigDict(validate_assignment=True, arbitrary_types_allowed=True)

    @pydantic.model_validator(mode='after')
    def ensure_some(self) -> Self:
        """
        Make sure that at least one of the dataframes is not `None` when the model is
        instantiated.
        """
        if self.df is None and self.pl_df is None and self.lf is None:
            raise AssertionError('At least one of df, pl_df, or lf must be set')
        return self

    @property
    def pandas(self) -> pd.DataFrame:
        """Return the pandas DataFrame, instantiating it if necessary."""
        if self.df is not None:
            return self.df

        if self.pl_df is not None:
            self.df = self.pl_df.to_pandas(use_pyarrow_extension_array=False)
            self.df[list(self.columns_with_iterables)] = self.df[
                list(self.columns_with_iterables)
            ].map(tuple)
            return self.df

        self.pl_df = self.lf.collect()  # type: ignore[union-attr]
        self.df = self.pl_df.to_pandas(use_pyarrow_extension_array=False)
        for colname in self.columns_with_iterables:
            self.df[colname] = self.df[colname].apply(tuple)
        return self.df

    @property
    def polars(self) -> pl.DataFrame:
        """Return the polars DataFrame, instantiating it if necessary."""
        if self.pl_df is not None:
            return self.pl_df

        if self.lf is not None:
            self.pl_df = self.lf.collect()
            return self.pl_df

        self.pl_df = pl.from_pandas(self.df)
        self.lf = self.pl_df.lazy()

        return self.pl_df

    @property
    def lazy(self) -> pl.LazyFrame:
        """Return the polars LazyFrame, instantiating it if necessary."""
        if self.lf is not None:
            return self.lf

        # Otherwise, it must be none - so lets create the lazyframe now. We use the
        # self.polars property, so we can cascade to creating it from the pandas dataframe
        # if necessary.
        self.lf = self.polars.lazy()
        return self.lf

    @property
    def columns_with_iterables(self) -> set[str]:
        """Return a set of columns that have iterables, preferentially using
        `self.lazy` > `self.polars` > `self.pandas` to minimise overhead."""
        if (trunc_df := self.lazy.head(1).collect()).is_empty():
            return set()
        if self.df is not None and self.df.empty:
            return set()

        colnames, dtypes = trunc_df.columns, trunc_df.dtypes
        return {colname for colname, dtype in zip(colnames, dtypes) if dtype == pl.List}

    def nunique(self) -> pd.Series:
        """Return a series of the number of unique values for each column in the catalog."""
        return pd.Series(
            {
                colname: self.polars.get_column(colname).explode().n_unique()
                if self.polars.schema[colname] == pl.List
                else self.polars.get_column(colname).n_unique()
                for colname in self.polars.columns
            }
        )


class CatalogFileDataReader:
    """Abstracts away some of the complexity related to reading dataframes"""

    def __init__(
        self,
        catalog_file: pydantic.StrictStr | None,
        storage_options: dict[str, typing.Any],
        **read_kwargs,
    ):
        self.catalog_file = catalog_file
        self.storage_options = storage_options
        self.read_kwargs = read_kwargs

        if self.catalog_file is None:
            raise AssertionError('catalog_file must be set to a valid file path or URL')

        # I think we want to replace this with a dict lookup.
        if self.catalog_file.endswith('.csv.gz') or self.catalog_file.endswith('.csv'):
            self.driver = 'polars'
            self.filetype = 'csv'
        elif self.catalog_file.endswith('.parquet'):
            self.driver = 'polars'
            self.filetype = 'parquet'
        elif self.catalog_file.endswith('.csv.bz2') or self.catalog_file.endswith('.csv.xz'):
            self.driver = 'pandas'
            self.filetype = 'csv'
        else:
            raise ValueError(
                f'Unsupported file type for catalog_file {self.catalog_file}. '
                f'Expected one of {__filetypes__}'
            )

        self._dtype_map: dict[str, str] = {}
        self.frames = self._read()

    def _read_csv_pd(self) -> FramesModel:
        """Read a catalog file stored as a csv using pandas"""
        df = pd.read_csv(
            self.catalog_file,
            storage_options=self.storage_options,
            **self.read_kwargs,
        )
        self._dtype_map = {
            colname: df['colname'].dtype
            for colname in self.read_kwargs.get('converters', {}).keys()
        }
        return FramesModel(df=df)

    def _read_csv_pl(self) -> FramesModel:
        """Read a catalog file stored as a csv using polars"""
        converters = self.read_kwargs.pop('converters', {})  # Hack
        # See https://github.com/pola-rs/polars/issues/13040 - can't use read_csv.
        lf = pl.scan_csv(
            self.catalog_file,  # type: ignore[arg-type]
            storage_options=self.storage_options,
            infer_schema=False,
            **self.read_kwargs,
        )

        if dtype_map := (
            lf.head(1)
            .select([colname for colname in converters.keys()])
            .with_columns(
                [
                    pl.col(colname)
                    .str.head(1)
                    .str.replace_many(
                        ['[', '(', '{'],
                        ['list', 'tuple', 'set'],
                    )
                    for colname in converters.keys()
                ]
            )
            .collect()
            .to_dicts()
        ):  # Returns an empty list if no rows - hence walrus
            self._dtype_map = dtype_map[0]

        lf = lf.with_columns(
            [
                pl.col(colname)
                .str.replace('^.', '[')  # Replace first/last chars with [ or ].
                .str.replace('.$', ']')  # set/tuple => list
                .str.replace(',]$', ']')  # Remove trailing commas
                .str.replace_all(
                    "'", '"'
                )  # This is to do with the JSON spec - single versus double quotes
                .str.json_decode(dtype=pl.List(pl.Utf8))
                for colname in converters.keys()
            ]
        )
        return FramesModel(lf=lf)

    def _read_parquet_pl(self) -> FramesModel:
        """Read a catalog file stored as a parquet using polars"""
        lf = pl.scan_parquet(
            self.catalog_file,  # type: ignore[arg-type]
            storage_options=self.storage_options,
            **self.read_kwargs,
        )
        return FramesModel(lf=lf)

    def _read(self):
        if self.driver == 'polars':
            if self.filetype == 'csv':
                return self._read_csv_pl()
            elif self.filetype == 'parquet':
                return self._read_parquet_pl()
            else:
                raise ValueError(f'Unsupported file type {self.filetype} for polars reader')

        if self.driver == 'pandas':
            if self.filetype == 'csv':
                return self._read_csv_pd()
            else:
                raise ValueError(f'Unsupported file type {self.filetype} for pandas reader')

    @property
    def dtype_map(self) -> dict[str, str]:
        """Return a map of column names to their dtypes for columns with iterables."""
        return self._dtype_map



import typing
import warnings

import dask
import fsspec
import pandas as pd
import pydantic
import xarray as xr
from intake.source.base import DataSource, Schema

from .cat import Aggregation, DataFormat
from .utils import OPTIONS, _set_async_flag


class ConcatenationWarning(UserWarning):
    pass


class ESMDataSourceError(Exception):
    pass


def _get_xarray_open_kwargs(data_format, xarray_open_kwargs=None, storage_options=None):
    xarray_open_kwargs = (xarray_open_kwargs or {}).copy()
    _default_open_kwargs = {
        'engine': 'zarr' if data_format in {'zarr', 'zarr2', 'zarr3', 'reference'} else 'netcdf4',
        'chunks': {},
        'backend_kwargs': {},
        'decode_timedelta': False,
    }
    xarray_open_kwargs = (
        {**_default_open_kwargs, **xarray_open_kwargs}
        if xarray_open_kwargs
        else _default_open_kwargs
    )

    if (
        xarray_open_kwargs['engine'] == 'zarr'
        and 'storage_options' not in xarray_open_kwargs['backend_kwargs']
    ):
        xarray_open_kwargs['backend_kwargs']['storage_options'] = {} or storage_options

    xarray_open_kwargs = _set_async_flag(data_format, xarray_open_kwargs)

    return xarray_open_kwargs


def _get_open_func(threaded: bool) -> typing.Callable:
    """Return the appropriate open function based on threading"""
    if threaded:
        return _delayed_open_ds
    else:
        return _eager_open_ds


def _eager_open_ds(*args, **kwargs):
    return _open_dataset(*args, **kwargs)


@dask.delayed
def _delayed_open_ds(*args, **kwargs):
    return _open_dataset(*args, **kwargs)


def _open_dataset(
    urlpath,
    varname,
    *,
    xarray_open_kwargs=None,
    preprocess=None,
    requested_variables=None,
    additional_attrs=None,
    expand_dims=None,
    data_format=None,
    storage_options=None,
):
    storage_options = storage_options or xarray_open_kwargs.get('backend_kwargs', {}).get(
        'storage_options', {}
    )

    # Support kerchunk datasets, setting the file object (fo) and urlpath
    if data_format == 'reference':
        xarray_open_kwargs['backend_kwargs']['storage_options']['fo'] = urlpath
        xarray_open_kwargs['backend_kwargs']['consolidated'] = False
        urlpath = 'reference://'

    if xarray_open_kwargs['engine'] in 'zarr' or data_format == 'opendap':
        url = urlpath
    elif fsspec.utils.can_be_local(urlpath):
        url = fsspec.open_local(urlpath, **storage_options)
    else:
        url = fsspec.open(urlpath, **storage_options).open()

    # Handle multi-file datasets with `xr.open_mfdataset()`
    if (isinstance(url, str) and '*' in url) or isinstance(url, list):
        # How should we handle concat_dim, and other xr.open_mfdataset kwargs?
        xarray_open_kwargs.update(preprocess=preprocess)
        xarray_open_kwargs.update(parallel=True)
        ds = xr.open_mfdataset(url, **xarray_open_kwargs)
    else:
        ds = xr.open_dataset(url, **xarray_open_kwargs)
        if preprocess is not None:
            ds = preprocess(ds)

    if varname and isinstance(varname, str):
        varname = [varname]

    if requested_variables:
        if isinstance(requested_variables, str):
            requested_variables = [requested_variables]

        variable_intersection = set(requested_variables).intersection(set(varname))

        data_vars = variable_intersection & set(ds.data_vars)
        coord_vars = variable_intersection & set(ds.coords)

        variables = list(data_vars | coord_vars)

        scalar_variables = [v for v in ds.data_vars if len(ds[v].dims) == 0]

        ds = ds.set_coords(scalar_variables)
        ds = ds[variables]
        ds.attrs[OPTIONS['vars_key']] = variables
    elif varname:
        ds.attrs[OPTIONS['vars_key']] = varname

    ds = _expand_dims(expand_dims, ds)
    ds = _update_attrs(additional_attrs=additional_attrs, ds=ds)
    return ds


def _update_attrs(*, additional_attrs, ds):
    additional_attrs = additional_attrs or {}
    if additional_attrs:
        additional_attrs = {
            f'{OPTIONS["attrs_prefix"]}:{key}': f'{value}'
            if isinstance(value, str) or not hasattr(value, '__iter__')
            else ','.join(value)
            for key, value in additional_attrs.items()
        }
    ds.attrs = {**ds.attrs, **additional_attrs}
    return ds


def _expand_dims(expand_dims, ds):
    if expand_dims:
        for variable in ds.attrs[OPTIONS['vars_key']]:
            ds[variable] = ds[variable].expand_dims(**expand_dims)

    return ds


class ESMDataSource(DataSource):
    version = '1.0'
    container = 'xarray'
    name = 'esm_datasource'
    partition_access = True

    @pydantic.validate_call
    def __init__(
        self,
        key: pydantic.StrictStr,
        records: list[dict[str, typing.Any]],
        path_column_name: pydantic.StrictStr,
        data_format: DataFormat | None,
        format_column_name: pydantic.StrictStr | None,
        *,
        variable_column_name: pydantic.StrictStr | None = None,
        aggregations: list[Aggregation] | None = None,
        requested_variables: list[str] | None = None,
        preprocess: typing.Callable | None = None,
        storage_options: dict[str, typing.Any] | None = None,
        xarray_open_kwargs: dict[str, typing.Any] | None = None,
        xarray_combine_by_coords_kwargs: dict[str, typing.Any] | None = None,
        intake_kwargs: dict[str, typing.Any] | None = None,
        threaded: bool,
    ):
        """An intake compatible Data Source for ESM data.

        Parameters
        ----------
        key: str
            The key of the data source.
        records: list of dict
            A list of records, each of which is a dictionary
            mapping column names to values.
        path_column_name: str
            The column name of the path.
        data_format: DataFormat
            The data format of the data.
        variable_column_name: str, optional
            The column name of the variable name.
        aggregations: list of Aggregation, optional
            A list of aggregations to apply to the data.
        requested_variables: list of str, optional
            A list of variables to load.
        preprocess: callable, optional
            A preprocessing function to apply to the data.
        storage_options: dict, optional
            fsspec parameters passed to the backend file-system such as Google Cloud Storage,
            Amazon Web Service S3.
        xarray_open_kwargs: dict, optional
            Keyword arguments to pass to :py:func:`~xarray.open_dataset` function.
        xarray_combine_by_coords_kwargs: dict, optional
            Keyword arguments to pass to :py:func:`~xarray.combine_by_coords` function.
        intake_kwargs: dict, optional
            Additional keyword arguments are passed through to the :py:class:`~intake.source.base.DataSource` base class.
        threaded : bool , optional
            If True, use `dask.compute` to load datasets in parallel. If False, load datasets sequentially.
            If none, the environment variable `ITK_ESM_THREADING` will be used to determine the threading behavior,
            defaulting to True if the variable is not set.
        """

        intake_kwargs = intake_kwargs or {}
        super().__init__(**intake_kwargs)
        self.key = key
        self.storage_options = storage_options or {}
        self.preprocess = preprocess
        self.requested_variables = requested_variables or []
        self.path_column_name = path_column_name
        self.variable_column_name = variable_column_name
        self.aggregations = aggregations
        self.df = pd.DataFrame.from_records(records)
        self.xarray_open_kwargs = xarray_open_kwargs
        self.xarray_combine_by_coords_kwargs = dict(combine_attrs='drop_conflicts')
        if xarray_combine_by_coords_kwargs is None:
            xarray_combine_by_coords_kwargs = {}
        self.xarray_combine_by_coords_kwargs = {
            **self.xarray_combine_by_coords_kwargs,
            **xarray_combine_by_coords_kwargs,
        }
        self.threaded = threaded
        self._ds = None

        if data_format is not None:
            self.df['_data_format_'] = data_format.value
        else:
            self.df = self.df.rename(columns={format_column_name: '_data_format_'})

    def __repr__(self) -> str:
        return f'<{type(self).__name__}  (name: {self.key}, asset(s): {len(self.df)})>'

    def _get_schema(self) -> Schema:
        if self._ds is None:
            self._open_dataset()
            metadata: dict[str, typing.Any] = {'dims': {}, 'data_vars': {}, 'coords': ()}
            self._schema = Schema(
                datashape=None,
                dtype=None,
                shape=None,
                npartitions=None,
                extra_metadata=metadata,
            )
        return self._schema

    def _open_dataset(self):
        """Open dataset with xarray"""

        open_ds_func = _get_open_func(self.threaded)

        try:
            datasets = [
                open_ds_func(
                    record[self.path_column_name],
                    record[self.variable_column_name] if self.variable_column_name else None,
                    xarray_open_kwargs=_get_xarray_open_kwargs(
                        record['_data_format_'], self.xarray_open_kwargs, self.storage_options
                    ),
                    preprocess=self.preprocess,
                    expand_dims={
                        agg.attribute_name: [record[agg.attribute_name]]
                        for agg in self.aggregations
                        if agg.type.value == 'join_new'
                    },
                    requested_variables=self.requested_variables,
                    data_format=record['_data_format_'],
                    additional_attrs=record[~record.isnull()].to_dict(),
                    storage_options=self.storage_options,
                )
                for _, record in self.df.iterrows()
            ]

            if self.threaded:
                datasets = dask.compute(*datasets)
            if len(datasets) == 1 or not datasets[0].data_vars:
                self._ds = datasets[0]
            else:
                datasets = sorted(
                    datasets,
                    key=lambda ds: tuple(
                        f'{OPTIONS["attrs_prefix"]}/{agg.attribute_name}'
                        for agg in self.aggregations
                    ),
                )
                datasets = [
                    ds.set_coords(set(ds.variables) - set(ds.attrs[OPTIONS['vars_key']]))
                    for ds in datasets
                ]
                try:
                    self._ds = xr.combine_by_coords(
                        datasets, **self.xarray_combine_by_coords_kwargs
                    )
                except ValueError as exc:
                    if (
                        str(exc)
                        == 'Could not find any dimension coordinates to use to order the datasets for concatenation'
                    ):
                        warnings.warn(
                            'Attempting to concatenate datasets without valid dimension coordinates: retaining only first dataset.'
                            ' Request valid dimension coordinate to silence this warning.',
                            category=ConcatenationWarning,
                        )
                        self._ds = datasets[0]
                    else:
                        raise exc

            self._ds.attrs[OPTIONS['dataset_key']] = self.key

        except Exception as exc:
            raise ESMDataSourceError(
                f"""Failed to load dataset with key='{self.key}'
                 You can use `cat['{self.key}'].df` to inspect the assets/files for this key.
                 """
            ) from exc

    def to_dask(self):
        """Return xarray object (which will have chunks)"""
        self._load_metadata()
        return self._ds

    def close(self):
        """Delete open files from memory"""
        self._ds = None
        self._schema = None



import ast
import concurrent.futures
import os
import typing
import warnings
from copy import deepcopy

import dask
import packaging.version
import xarray as xr

try:
    if packaging.version.Version(xr.__version__) < packaging.version.Version('2024.10'):
        from datatree import DataTree
    else:
        from xarray import DataTree
    _DATATREE_AVAILABLE = True
except ImportError:
    _DATATREE_AVAILABLE = False
import itables
import pandas as pd
import polars as pl
import pydantic
from fastprogress.fastprogress import progress_bar
from intake.catalog import Catalog

from .cat import ESMCatalogModel
from .derived import DerivedVariableRegistry, default_registry
from .source import ESMDataSource
from .utils import MinimalExploder


class esm_datastore(Catalog):
    """
    An intake plugin for parsing an ESM (Earth System Model) Catalog
    and loading assets (netCDF files and/or Zarr stores) into xarray datasets.
    The in-memory representation for the catalog is a Pandas DataFrame.

    Parameters
    ----------
    obj : str, dict, ESMCatalogModel
        The ESM Catalog to use, or a path to a JSON file containing the catalog.
        If string, this must be a path or URL to an ESM catalog JSON file.
        If dict, this must be a dict representation of an ESM catalog.
        This dict must have two keys: 'esmcat' and 'df'. The 'esmcat' key must be a
        dict representation of the ESM catalog and the 'df' key must
        be a Pandas DataFrame containing content that would otherwise be in a CSV file.
    sep : str, optional
        Delimiter to use when constructing a key for a query, by default '.'
    registry : DerivedVariableRegistry, optional
        Registry of derived variables to use, by default None. If not provided, uses the default registry.
    read_kwargs : dict, optional
        Additional keyword arguments passed through to the :py:func:`~polars.scan_csv` function, if the
        datastore is saved in csv format, or :py:func:`~polars.scan_parquet` if the datastore is saved in
        parquet format.
    read_csv_kwargs : dict, optional
        Deprecated alias for `read_kwargs`.
    columns_with_iterables : list of str, optional
        A list of columns in the csv file containing iterables. Values in columns specified here will be
        converted with `ast.literal_eval` when :py:func:`~pandas.read_csv` is called (i.e., this is a
        shortcut to passing converters to `read_kwargs`).
    storage_options : dict, optional
        Parameters passed to the backend file-system such as Google Cloud Storage,
        Amazon Web Service S3.
    intake_kwargs: dict, optional
        Additional keyword arguments are passed through to the :py:class:`~intake.catalog.Catalog` base class.

    Examples
    --------

    At import time, this plugin is available in intake's registry as `esm_datastore` and
    can be accessed with `intake.open_esm_datastore()`:

    >>> import intake
    >>> url = 'https://storage.googleapis.com/cmip6/pangeo-cmip6.json'
    >>> cat = intake.open_esm_datastore(url)
    >>> cat.df.head()
    activity_id institution_id source_id experiment_id  ... variable_id grid_label                                             zstore dcpp_init_year
    0  AerChemMIP            BCC  BCC-ESM1        ssp370  ...          pr         gn  gs://cmip6/AerChemMIP/BCC/BCC-ESM1/ssp370/r1i1...            NaN
    1  AerChemMIP            BCC  BCC-ESM1        ssp370  ...        prsn         gn  gs://cmip6/AerChemMIP/BCC/BCC-ESM1/ssp370/r1i1...            NaN
    2  AerChemMIP            BCC  BCC-ESM1        ssp370  ...         tas         gn  gs://cmip6/AerChemMIP/BCC/BCC-ESM1/ssp370/r1i1...            NaN
    3  AerChemMIP            BCC  BCC-ESM1        ssp370  ...      tasmax         gn  gs://cmip6/AerChemMIP/BCC/BCC-ESM1/ssp370/r1i1...            NaN
    4  AerChemMIP            BCC  BCC-ESM1        ssp370  ...      tasmin         gn  gs://cmip6/AerChemMIP/BCC/BCC-ESM1/ssp370/r1i1...            NaN
    """

    name = 'esm_datastore'
    container = 'xarray'

    def __init__(
        self,
        obj: pydantic.FilePath | pydantic.AnyUrl | dict[str, typing.Any] | ESMCatalogModel,
        *,
        progressbar: bool = True,
        sep: str = '.',
        registry: DerivedVariableRegistry | None = None,
        read_kwargs: dict[str, typing.Any] | None = None,
        read_csv_kwargs: dict[str, typing.Any] | None = None,
        columns_with_iterables: list[str] | None = None,
        storage_options: dict[str, typing.Any] | None = None,
        threaded: bool | None = None,
        **intake_kwargs: dict[str, typing.Any],
    ):
        """Intake Catalog representing an ESM Collection."""
        super().__init__(**intake_kwargs)
        self.storage_options = storage_options or {}

        if read_csv_kwargs is not None:
            warnings.warn(
                'read_csv_kwargs is deprecated and will be removed in a future version. '
                'Please use read_kwargs instead.',
                DeprecationWarning,
                stacklevel=2,
            )
            if read_kwargs is not None:
                raise ValueError(
                    'Cannot provide both `read_csv_kwargs` and `read_kwargs`. '
                    'Please use `read_kwargs`.'
                )
            read_kwargs = read_csv_kwargs

        read_kwargs = read_kwargs or {}
        if columns_with_iterables:
            converter = ast.literal_eval
            read_kwargs.setdefault('converters', {})
            for col in columns_with_iterables:
                if read_kwargs['converters'].setdefault(col, converter) != converter:
                    raise ValueError(
                        f"Cannot provide converter for '{col}' via `read_kwargs` when '{col}' is also specified in `columns_with_iterables`"
                    )
        self.read_kwargs = read_kwargs
        self.progressbar = progressbar
        self.sep = sep

        if threaded is None:
            self.threaded = ast.literal_eval(os.getenv('ITK_ESM_THREADING', 'True'))
        else:
            self.threaded = threaded

        if isinstance(obj, ESMCatalogModel):
            self.esmcat = obj
        elif isinstance(obj, dict):
            self.esmcat = ESMCatalogModel.from_dict(obj)
        else:
            self.esmcat = ESMCatalogModel.load(
                obj, storage_options=self.storage_options, read_kwargs=read_kwargs
            )

        self.derivedcat = registry or default_registry
        self._entries = {}
        self._requested_variables = []
        self._columns_with_iterables = columns_with_iterables or []
        self.datasets = {}
        self._validate_derivedcat()

    def _validate_derivedcat(self) -> None:
        if self.esmcat.aggregation_control is None and len(self.derivedcat):
            raise ValueError(
                'Variable derivation requires `aggregation_control` to be specified in the catalog.'
            )
        for key, entry in self.derivedcat.items():
            if self.esmcat.aggregation_control.variable_column_name not in entry.query.keys():
                raise ValueError(
                    f'Variable derivation requires `{self.esmcat.aggregation_control.variable_column_name}` to be specified in query: {entry.query} for derived variable {key}.'
                )

            for col in entry.query:
                if col not in self.esmcat.df.columns:
                    raise ValueError(
                        f'Derived variable {key} depends on unknown column {col} in query: {entry.query}. Valid ESM catalog columns: {self.esmcat.df.columns.tolist()}.'
                    )

    def keys(self) -> list[str]:
        """
        Get keys for the catalog entries

        Returns
        -------
        list
            keys for the catalog entries
        """
        return list(self.esmcat._construct_group_keys(sep=self.sep).keys())

    def keys_info(self) -> pd.DataFrame:
        """
        Get keys for the catalog entries and their metadata

        Returns
        -------
        pandas.DataFrame
            keys for the catalog entries and their metadata

        Examples
        --------

        >>> import intake
        >>> cat = intake.open_esm_datastore('./tests/sample-catalogs/cesm1-lens-netcdf.json')
        >>> cat.keys_info()
                        component experiment stream
        key
        ocn.20C.pop.h         ocn        20C  pop.h
        ocn.CTRL.pop.h        ocn       CTRL  pop.h
        ocn.RCP85.pop.h       ocn      RCP85  pop.h



        """
        results = self.esmcat._construct_group_keys(sep=self.sep)
        if self.esmcat.aggregation_control and self.esmcat.aggregation_control.groupby_attrs:
            groupby_attrs = self.esmcat.aggregation_control.groupby_attrs
        else:
            groupby_attrs = list(self.df.columns)
        data = {key: dict(zip(groupby_attrs, results[key])) for key in results}
        data = pd.DataFrame.from_dict(data, orient='index')
        data.index.name = 'key'
        return data

    @property
    def key_template(self) -> str:
        """
        Return string template used to create catalog entry keys

        Returns
        -------
        str
          string template used to create catalog entry keys
        """
        if self.esmcat.aggregation_control and self.esmcat.aggregation_control.groupby_attrs:
            return self.sep.join(self.esmcat.aggregation_control.groupby_attrs)
        else:
            return self.sep.join(self.esmcat.df.columns)

    @property
    def df(self) -> pd.DataFrame:
        """
        Return pandas :py:class:`~pandas.DataFrame`.
        """
        return self.esmcat.df

    @property
    def interactive(self) -> None:
        """
        Use itables to display the catalog in an interactive table. Use polars
        for performance ideally. Fall back to pandas if not.

        We have to explode columns with iterables, otherwise javascript stringifcation
        can cause ellipsis to be rendered directly into the interactive table,
        losing actual data and inserting junk.
        """

        try:
            pl_df = self.esmcat._frames.polars  # type:ignore[union-attr]
        except AttributeError:
            pl_df = pl.from_pandas(self.df)

        exploded_df = MinimalExploder(pl_df)()

        return itables.show(
            exploded_df,
            search={'regex': True, 'caseInsensitive': True},
            layout={'top1': 'searchPanes'},
            searchPanes={
                'layout': 'columns-3',
                'cascadePanes': True,
                'columns': [i for i, _ in enumerate(pl_df.columns)],
            },
            maxBytes=0,
        )

    def __len__(self) -> int:
        return len(self.keys())

    def _get_entries(self) -> dict[str, ESMDataSource]:
        # Due to just-in-time entry creation, we may not have all entries loaded
        # We need to make sure to create entries missing from self._entries
        missing = set(self.keys()) - set(self._entries.keys())
        for key in missing:
            _ = self[key]
        return self._entries

    @pydantic.validate_call
    def __getitem__(self, key: str) -> ESMDataSource:
        """
        This method takes a key argument and return a data source
        corresponding to assets (files) that will be aggregated into a
        single xarray dataset.

        Parameters
        ----------
        key : str
          key to use for catalog entry lookup

        Returns
        -------
        intake_esm.source.ESMDataSource
             A data source by name (key)

        Raises
        ------
        KeyError
            if key is not found.

        Examples
        --------
        >>> cat = intake.open_esm_datastore('mycatalog.json')
        >>> data_source = cat['AerChemMIP.BCC.BCC-ESM1.piClim-control.AERmon.gn']
        """
        # The canonical unique key is the key of a compatible group of assets
        try:
            return self._entries[key]
        except KeyError as e:
            if key in self.keys():
                keys_dict = self.esmcat._construct_group_keys(sep=self.sep)
                grouped = self.esmcat.grouped

                internal_key = keys_dict[key]

                if isinstance(grouped, pd.DataFrame):
                    records = [grouped.loc[internal_key].to_dict()]

                else:
                    records = grouped.get_group(internal_key).to_dict(orient='records')

                if self.esmcat.aggregation_control:
                    variable_column_name = self.esmcat.aggregation_control.variable_column_name
                    aggregations = self.esmcat.aggregation_control.aggregations
                else:
                    variable_column_name = None
                    aggregations = []
                # Create a new entry
                entry = ESMDataSource(
                    key=key,
                    records=records,
                    variable_column_name=variable_column_name,
                    path_column_name=self.esmcat.assets.column_name,
                    data_format=self.esmcat.assets.format,
                    format_column_name=self.esmcat.assets.format_column_name,
                    aggregations=aggregations,
                    intake_kwargs={'metadata': {}},
                    threaded=self.threaded,
                )
                self._entries[key] = entry
                return self._entries[key]
            raise KeyError(
                f'key={key} not found in catalog. You can access the list of valid keys via the .keys() method.'
            ) from e

    def __contains__(self, key) -> bool:
        # Python falls back to iterating over the entire catalog
        # if this method is not defined. To avoid this, we implement it differently

        try:
            self[key]
        except KeyError:
            return False
        else:
            return True

    def __repr__(self) -> str:
        """Make string representation of object."""
        return f'<{self.esmcat.id or ""} catalog with {len(self)} dataset(s) from {len(self.df)} asset(s)>'

    def _repr_html_(self) -> str:
        """
        Return an html representation for the catalog object.
        Mainly for IPython notebook
        """
        uniques = pd.DataFrame(self.nunique(), columns=['unique'])
        text = uniques._repr_html_()
        return f'<p><strong>{self.esmcat.id or ""} catalog with {len(self)} dataset(s) from {len(self.df)} asset(s)</strong>:</p> {text}'

    def _ipython_display_(self):
        """
        Display the entry as a rich object in an IPython session
        """
        from IPython.display import HTML, display

        contents = self._repr_html_()
        display(HTML(contents))

    def __dir__(self) -> list[str]:
        rv = [
            'df',
            'to_dataset_dict',
            'to_datatree',
            'to_dask',
            'keys',
            'keys_info',
            'serialize',
            'datasets',
            'search',
            'unique',
            'nunique',
            'key_template',
        ]
        return sorted(list(self.__dict__.keys()) + rv)

    def _ipython_key_completions_(self):
        return self.__dir__()

    @pydantic.validate_call
    def search(
        self,
        require_all_on: str | list[str] | None = None,
        **query: typing.Any,
    ):
        """Search for entries in the catalog.

        Parameters
        ----------
        require_all_on : list, str, optional
            A dataframe column or a list of dataframe columns across
            which all entries must satisfy the query criteria.
            If None, return entries that fulfill any of the criteria specified
            in the query, by default None.
        **query:
            keyword arguments corresponding to user's query to execute against the dataframe.

        Returns
        -------
        cat : :py:class:`~intake_esm.core.esm_datastore`
          A new Catalog with a subset of the entries in this Catalog.

        Examples
        --------
        >>> import intake
        >>> cat = intake.open_esm_datastore('pangeo-cmip6.json')
        >>> cat.df.head(3)
        activity_id institution_id source_id  ... grid_label                                             zstore dcpp_init_year
        0  AerChemMIP            BCC  BCC-ESM1  ...         gn  gs://cmip6/AerChemMIP/BCC/BCC-ESM1/ssp370/r1i1...            NaN
        1  AerChemMIP            BCC  BCC-ESM1  ...         gn  gs://cmip6/AerChemMIP/BCC/BCC-ESM1/ssp370/r1i1...            NaN
        2  AerChemMIP            BCC  BCC-ESM1  ...         gn  gs://cmip6/AerChemMIP/BCC/BCC-ESM1/ssp370/r1i1...            NaN

        >>> sub_cat = cat.search(
        ...     source_id=['BCC-CSM2-MR', 'CNRM-CM6-1', 'CNRM-ESM2-1'],
        ...     experiment_id=['historical', 'ssp585'],
        ...     variable_id='pr',
        ...     table_id='Amon',
        ...     grid_label='gn',
        ... )
        >>> sub_cat.df.head(3)
            activity_id institution_id    source_id  ... grid_label                                             zstore dcpp_init_year
        260        CMIP            BCC  BCC-CSM2-MR  ...         gn  gs://cmip6/CMIP/BCC/BCC-CSM2-MR/historical/r1i...            NaN
        346        CMIP            BCC  BCC-CSM2-MR  ...         gn  gs://cmip6/CMIP/BCC/BCC-CSM2-MR/historical/r2i...            NaN
        401        CMIP            BCC  BCC-CSM2-MR  ...         gn  gs://cmip6/CMIP/BCC/BCC-CSM2-MR/historical/r3i...            NaN

        The search method also accepts compiled regular expression objects
        from :py:func:`~re.compile` as patterns.

        >>> import re
        >>> # Let's search for variables containing "Frac" in their name
        >>> pat = re.compile(r'Frac')  # Define a regular expression
        >>> cat.search(variable_id=pat)
        >>> cat.df.head().variable_id
        0     residualFrac
        1    landCoverFrac
        2    landCoverFrac
        3     residualFrac
        4    landCoverFrac
        """

        # step 1: Search in the base/main catalog
        esmcat_results = self.esmcat.search(require_all_on=require_all_on, query=query)

        # step 2: Search for entries required to derive variables in the derived catalogs
        # This requires a bit of a hack i.e. the user has to specify the variable in the query
        derivedcat_results = []
        if self.esmcat.aggregation_control:
            variables = query.pop(self.esmcat.aggregation_control.variable_column_name, None)
        else:
            variables = None
        dependents = []
        derived_cat_subset = {}
        if variables:
            if isinstance(variables, str):
                variables = [variables]
            for key, value in self.derivedcat.items():
                if key in variables:
                    res = self.esmcat.search(
                        require_all_on=require_all_on, query={**value.query, **query}
                    )
                    if not res.empty:
                        derivedcat_results.append(res)
                        dependents.extend(
                            value.dependent_variables(
                                self.esmcat.aggregation_control.variable_column_name
                            )
                        )
                        derived_cat_subset[key] = value

        if derivedcat_results:
            # Merge results from the main and the derived catalogs
            esmcat_results = pd.concat([esmcat_results, *derivedcat_results])
            esmcat_results = esmcat_results[~esmcat_results.astype(str).duplicated()]

        cat = self.__class__({'esmcat': self.esmcat.model_dump(), 'df': esmcat_results})
        cat.esmcat.catalog_file = None  # Don't save the catalog file
        if self.esmcat.has_multiple_variable_assets:
            requested_variables = list(set(variables or []).union(dependents))
        else:
            requested_variables = []
        cat._requested_variables = requested_variables

        # step 3: Subset the derived catalog,
        # but only if variables were looked up, otherwise transfer the whole catalog.
        if variables is not None:
            cat.derivedcat = DerivedVariableRegistry()
            cat.derivedcat._registry.update(derived_cat_subset)
        else:
            cat.derivedcat = self.derivedcat
        return cat

    @pydantic.validate_call
    def serialize(
        self,
        name: pydantic.StrictStr,
        directory: pydantic.DirectoryPath | pydantic.StrictStr | None = None,
        catalog_type: str = 'dict',
        file_format: str = 'csv',
        write_kwargs: dict[typing.Any, typing.Any] | None = None,
        to_csv_kwargs: dict[typing.Any, typing.Any] | None = None,
        json_dump_kwargs: dict[typing.Any, typing.Any] | None = None,
        storage_options: dict[str, typing.Any] | None = None,
    ) -> None:
        """Serialize catalog to corresponding json and csv (or parquet) files.

        Parameters
        ----------
        name : str
            name to use when creating ESM catalog json file and csv catalog.
        directory : str, PathLike, default None
            The path to the local directory. If None, use the current directory
        catalog_type: str, default 'dict'
            Whether to save the catalog table as a dictionary in the JSON file or as a separate CSV file.
        file_format: str, default 'csv'
            The file format to use when saving the catalog table. Either 'csv' or 'parquet'.
            If catalog_type is 'dict', this parameter is ignored.
        write_kwargs: dict, optional
            Additional keyword arguments passed through to the :py:func:`~pandas.DataFrame.to_csv` or
            :py:func:`~pandas.DataFrame.to_parquet` functions.
            Compression is currently ignored when serializing to parquet.
        to_csv_kwargs : dict, optional
            Deprecated alias for `write_kwargs`.
        json_dump_kwargs : dict, optional
            Additional keyword arguments passed through to the :py:func:`~json.dump` function.
        storage_options: dict
            fsspec parameters passed to the backend file-system such as Google Cloud Storage,
            Amazon Web Service S3.

        Notes
        -----
        Large catalogs can result in large JSON files. To keep the JSON file size manageable, call with
        `catalog_type='file'` to save catalog as a separate CSV/parquet file.

        Examples
        --------
        >>> import intake
        >>> cat = intake.open_esm_datastore('pangeo-cmip6.json')
        >>> cat_subset = cat.search(
        ...     source_id='BCC-ESM1',
        ...     grid_label='gn',
        ...     table_id='Amon',
        ...     experiment_id='historical',
        ... )
        >>> cat_subset.serialize(name='cmip6_bcc_esm1', catalog_type='file')
        """

        if to_csv_kwargs is not None:
            warnings.warn(
                'to_csv_kwargs is deprecated and will be removed in a future version. '
                'Please use write_kwargs instead.',
                DeprecationWarning,
                stacklevel=2,
            )
            if write_kwargs is not None:
                raise ValueError(
                    'Cannot provide both `to_csv_kwargs` and `write_kwargs`. '
                    'Please use `write_kwargs`.'
                )
            write_kwargs = to_csv_kwargs

        write_kwargs = write_kwargs or {}

        self.esmcat.save(
            name,
            directory=directory,
            catalog_type=catalog_type,
            file_format=file_format,
            write_kwargs=write_kwargs,
            json_dump_kwargs=json_dump_kwargs,
            storage_options=storage_options,
        )

    def nunique(self) -> pd.Series:
        """Count distinct observations across dataframe columns
        in the catalog.

        Examples
        --------
        >>> import intake
        >>> cat = intake.open_esm_datastore('pangeo-cmip6.json')
        >>> cat.nunique()
        activity_id          10
        institution_id       23
        source_id            48
        experiment_id        29
        member_id            86
        table_id             19
        variable_id         187
        grid_label            7
        zstore            27437
        dcpp_init_year       59
        dtype: int64
        """
        nunique = self.esmcat.nunique()
        if self.esmcat.aggregation_control:
            nunique[f'derived_{self.esmcat.aggregation_control.variable_column_name}'] = len(
                self.derivedcat.keys()
            )
        return nunique

    def unique(self) -> pd.Series:
        """Return unique values for given columns in the
        catalog.
        """
        unique = self.esmcat.unique()
        if self.esmcat.aggregation_control:
            unique[f'derived_{self.esmcat.aggregation_control.variable_column_name}'] = list(
                self.derivedcat.keys()
            )
        return unique

    @pydantic.validate_call
    def to_dataset_dict(
        self,
        xarray_open_kwargs: dict[str, typing.Any] | None = None,
        xarray_combine_by_coords_kwargs: dict[str, typing.Any] | None = None,
        preprocess: typing.Callable | None = None,
        storage_options: dict[pydantic.StrictStr, typing.Any] | None = None,
        progressbar: pydantic.StrictBool | None = None,
        aggregate: pydantic.StrictBool | None = None,
        skip_on_error: pydantic.StrictBool = False,
        threaded: bool | None = None,
        **kwargs,
    ) -> dict[str, xr.Dataset]:
        """
        Load catalog entries into a dictionary of xarray datasets.

        Column values, dataset keys and requested variables are added as global
        attributes on the returned datasets. The names of these attributes can be
        customized with :py:class:`intake_esm.utils.set_options`.

        Parameters
        ----------
        xarray_open_kwargs : dict
            Keyword arguments to pass to :py:func:`~xarray.open_dataset` function
        xarray_combine_by_coords_kwargs: : dict
            Keyword arguments to pass to :py:func:`~xarray.combine_by_coords` function.
        preprocess : callable, optional
            If provided, call this function on each dataset prior to aggregation.
        storage_options : dict, optional
            fsspec Parameters passed to the backend file-system such as Google Cloud Storage,
            Amazon Web Service S3.
        progressbar : bool
            If True, will print a progress bar to standard error (stderr)
            when loading assets into :py:class:`~xarray.Dataset`.
        aggregate : bool, optional
            If False, no aggregation will be done.
        skip_on_error : bool, optional
            If True, skip datasets that cannot be loaded and/or variables we are unable to derive.
        threaded : bool, optional
            If True, use :py:func:`dask.compute` to load datasets in parallel. If False, load datasets sequentially.
            If None, the environment variable `ITK_ESM_THREADING` will be used to determine the threading behavior,
            defaulting to True if the variable is not set. If a value is provided, it will override the environment
            variable determined default.
        Returns
        -------
        dsets : dict
           A dictionary of xarray :py:class:`~xarray.Dataset`.

        Examples
        --------
        >>> import intake
        >>> cat = intake.open_esm_datastore('glade-cmip6.json')
        >>> sub_cat = cat.search(
        ...     source_id=['BCC-CSM2-MR', 'CNRM-CM6-1', 'CNRM-ESM2-1'],
        ...     experiment_id=['historical', 'ssp585'],
        ...     variable_id='pr',
        ...     table_id='Amon',
        ...     grid_label='gn',
        ... )
        >>> dsets = sub_cat.to_dataset_dict()
        >>> dsets.keys()
        dict_keys(['CMIP.BCC.BCC-CSM2-MR.historical.Amon.gn', 'ScenarioMIP.BCC.BCC-CSM2-MR.ssp585.Amon.gn'])
        >>> dsets['CMIP.BCC.BCC-CSM2-MR.historical.Amon.gn']
        <xarray.Dataset>
        Dimensions:    (bnds: 2, lat: 160, lon: 320, member_id: 3, time: 1980)
        Coordinates:
        * lon        (lon) float64 0.0 1.125 2.25 3.375 ... 355.5 356.6 357.8 358.9
        * lat        (lat) float64 -89.14 -88.03 -86.91 -85.79 ... 86.91 88.03 89.14
        * time       (time) object 1850-01-16 12:00:00 ... 2014-12-16 12:00:00
        * member_id  (member_id) <U8 'r1i1p1f1' 'r2i1p1f1' 'r3i1p1f1'
        Dimensions without coordinates: bnds
        Data variables:
            lat_bnds   (lat, bnds) float64 dask.array<chunksize=(160, 2), meta=np.ndarray>
            lon_bnds   (lon, bnds) float64 dask.array<chunksize=(320, 2), meta=np.ndarray>
            time_bnds  (time, bnds) object dask.array<chunksize=(1980, 2), meta=np.ndarray>
            pr         (member_id, time, lat, lon) float32 dask.array<chunksize=(1, 600, 160, 320), meta=np.ndarray>
        """

        # Return fast
        if not self.keys():
            warnings.warn(
                'There are no datasets to load! Returning an empty dictionary.',
                UserWarning,
                stacklevel=2,
            )
            return {}

        if (
            self.esmcat.aggregation_control
            and (
                self.esmcat.aggregation_control.variable_column_name
                in self.esmcat.aggregation_control.groupby_attrs
            )
            and len(self.derivedcat) > 0
        ):
            raise NotImplementedError(
                f'The `{self.esmcat.aggregation_control.variable_column_name}` column name is used as a groupby attribute: {self.esmcat.aggregation_control.groupby_attrs}. '
                'This is not yet supported when computing derived variables.'
            )

        threaded = _get_threaded(threaded)

        xarray_open_kwargs = xarray_open_kwargs or {}
        xarray_combine_by_coords_kwargs = xarray_combine_by_coords_kwargs or {}
        cdf_kwargs, zarr_kwargs = kwargs.get('cdf_kwargs'), kwargs.get('zarr_kwargs')

        if cdf_kwargs or zarr_kwargs:
            warnings.warn(
                'cdf_kwargs and zarr_kwargs are deprecated and will be removed in a future version. '
                'Please use xarray_open_kwargs instead.',
                DeprecationWarning,
                stacklevel=2,
            )
        if cdf_kwargs:
            xarray_open_kwargs.update(cdf_kwargs)
        if zarr_kwargs:
            xarray_open_kwargs.update(zarr_kwargs)

        source_kwargs = dict(
            xarray_open_kwargs=xarray_open_kwargs,
            xarray_combine_by_coords_kwargs=xarray_combine_by_coords_kwargs,
            preprocess=preprocess,
            requested_variables=self._requested_variables,
            storage_options=storage_options,
            threaded=threaded,
        )

        if aggregate is not None and not aggregate and self.esmcat.aggregation_control:
            self = deepcopy(self)
            self.esmcat.aggregation_control.groupby_attrs = []
        if progressbar is not None:
            self.progressbar = progressbar
        if self.progressbar:
            print(
                f"""\n--> The keys in the returned dictionary of datasets are constructed as follows:\n\t'{self.key_template}'"""
            )
        sources = {key: source(**source_kwargs) for key, source in self.items()}
        datasets = {}
        with concurrent.futures.ThreadPoolExecutor(max_workers=dask.system.CPU_COUNT) as executor:
            future_tasks = [
                executor.submit(_load_source, key, source) for key, source in sources.items()
            ]
            if self.progressbar:
                gen = progress_bar(
                    concurrent.futures.as_completed(future_tasks), total=len(sources)
                )
            else:
                gen = concurrent.futures.as_completed(future_tasks)
            for task in gen:
                try:
                    key, ds = task.result()
                    datasets[key] = ds
                except Exception as exc:
                    if not skip_on_error:
                        raise exc
        self.datasets = self._create_derived_variables(datasets, skip_on_error)
        return self.datasets

    @pydantic.validate_call
    def to_datatree(
        self,
        xarray_open_kwargs: dict[str, typing.Any] | None = None,
        xarray_combine_by_coords_kwargs: dict[str, typing.Any] | None = None,
        preprocess: typing.Callable | None = None,
        storage_options: dict[pydantic.StrictStr, typing.Any] | None = None,
        progressbar: pydantic.StrictBool | None = None,
        aggregate: pydantic.StrictBool | None = None,
        skip_on_error: pydantic.StrictBool = False,
        levels: list[str] = None,
        **kwargs,
    ):
        """
        Load catalog entries into a tree of xarray datasets.

        Parameters
        ----------
        xarray_open_kwargs : dict
            Keyword arguments to pass to :py:func:`~xarray.open_dataset` function
        xarray_combine_by_coords_kwargs: : dict
            Keyword arguments to pass to :py:func:`~xarray.combine_by_coords` function.
        preprocess : callable, optional
            If provided, call this function on each dataset prior to aggregation.
        storage_options : dict, optional
            Parameters passed to the backend file-system such as Google Cloud Storage,
            Amazon Web Service S3.
        progressbar : bool
            If True, will print a progress bar to standard error (stderr)
            when loading assets into :py:class:`~xarray.Dataset`.
        aggregate : bool, optional
            If False, no aggregation will be done.
        skip_on_error : bool, optional
            If True, skip datasets that cannot be loaded and/or variables we are unable to derive.
        levels : list[str], optional
            List of fields to use as the datatree nodes. WARNING: This will overwrite the fields
            used to create the unique aggregation keys.

        Returns
        -------
        dsets : :py:class:`~datatree.DataTree`
           A tree of xarray :py:class:`~xarray.Dataset`.

        Examples
        --------
        >>> import intake
        >>> cat = intake.open_esm_datastore('glade-cmip6.json')
        >>> sub_cat = cat.search(
        ...     source_id=['BCC-CSM2-MR', 'CNRM-CM6-1', 'CNRM-ESM2-1'],
        ...     experiment_id=['historical', 'ssp585'],
        ...     variable_id='pr',
        ...     table_id='Amon',
        ...     grid_label='gn',
        ... )
        >>> dsets = sub_cat.to_datatree()
        >>> dsets['CMIP/BCC.BCC-CSM2-MR/historical/Amon/gn'].ds
        <xarray.Dataset>
        Dimensions:    (bnds: 2, lat: 160, lon: 320, member_id: 3, time: 1980)
        Coordinates:
        * lon        (lon) float64 0.0 1.125 2.25 3.375 ... 355.5 356.6 357.8 358.9
        * lat        (lat) float64 -89.14 -88.03 -86.91 -85.79 ... 86.91 88.03 89.14
        * time       (time) object 1850-01-16 12:00:00 ... 2014-12-16 12:00:00
        * member_id  (member_id) <U8 'r1i1p1f1' 'r2i1p1f1' 'r3i1p1f1'
        Dimensions without coordinates: bnds
        Data variables:
            lat_bnds   (lat, bnds) float64 dask.array<chunksize=(160, 2), meta=np.ndarray>
            lon_bnds   (lon, bnds) float64 dask.array<chunksize=(320, 2), meta=np.ndarray>
            time_bnds  (time, bnds) object dask.array<chunksize=(1980, 2), meta=np.ndarray>
            pr         (member_id, time, lat, lon) float32 dask.array<chunksize=(1, 600, 160, 320), meta=np.ndarray>
        """

        if not _DATATREE_AVAILABLE:
            raise ImportError(
                '.to_datatree() requires the xarray-datatree package to be installed. '
                'To proceed please install xarray-datatree using: '
                ' `python -m pip install xarray-datatree` or `conda install -c conda-forge xarray-datatree`.'
            )

        # Change the groupby controls if neccessary, used to assemble the tree
        if levels is not None:
            self = deepcopy(self)
            self.esmcat.aggregation_control.groupby_attrs = levels

        # Set the separator to a / for datatree temporarily
        self.sep, old_sep = '/', self.sep

        # Use to dataset dict to access dictionary of datasets
        self.datasets = self.to_dataset_dict(
            xarray_open_kwargs=xarray_open_kwargs,
            xarray_combine_by_coords_kwargs=xarray_combine_by_coords_kwargs,
            preprocess=preprocess,
            storage_options=storage_options,
            progressbar=progressbar,
            aggregate=aggregate,
            skip_on_error=skip_on_error,
            **kwargs,
        )

        # Set the separator to the original value
        self.sep = old_sep

        # Convert the dictionary of datasets to a datatree
        self.datasets = DataTree.from_dict(self.datasets)
        return self.datasets

    def to_dask(self, **kwargs) -> xr.Dataset:
        """
        Convert result to an xarray dataset.

        This is only possible if the search returned exactly one result.

        Parameters
        ----------
        kwargs: dict
          Parameters forwarded to :py:func:`~intake_esm.esm_datastore.to_dataset_dict`.

        Returns
        -------
        :py:class:`~xarray.Dataset`
        """
        if len(self) != 1:  # quick check to fail more quickly if there are many results
            lens = self._aggr_lengths
            raise ValueError(
                f'Expected exactly one dataset. Received {len(self)} datasets. Please refine your search on {", ".join(lens.keys())} or use `.to_dataset_dict()`.'
            )
        res = self.to_dataset_dict(**{**kwargs, 'progressbar': False})
        if len(res) != 1:  # extra check in case kwargs did modify something
            lens = self._aggr_lengths
            raise ValueError(
                f'Expected exactly one dataset. Received {len(self)} datasets. Please refine your search on {", ".join(lens.keys())} or use `.to_dataset_dict()`.'
            )
        _, ds = res.popitem()
        return ds

    @property
    def _aggr_lengths(self) -> dict[str, int]:
        """
        For groupby attrs that have more than one unique value, return their name and length
        """
        lens = {
            k: len(getattr(self.unique(), k)) for k in self.esmcat.aggregation_control.groupby_attrs
        }
        return {k: v for k, v in lens.items() if v > 1}

    def _create_derived_variables(self, datasets, skip_on_error):
        if len(self.derivedcat) > 0:
            datasets = self.derivedcat.update_datasets(
                datasets=datasets,
                variable_key_name=self.esmcat.aggregation_control.variable_column_name,
                skip_on_error=skip_on_error,
            )
        return datasets


def _load_source(key, source):
    return key, source.to_dask()


def _get_threaded(threaded: bool | None) -> bool:
    """
    Read the threading option from the environment variable & passed value
    """
    if threaded is None:
        try:
            threaded = ast.literal_eval(os.getenv('ITK_ESM_THREADING', 'True'))
        except ValueError as e:
            raise ValueError(
                'The environment variable ITK_ESM_THREADING must be a boolean, if set.'
            ) from e

    return threaded



"""
Useful for:
* users learning intake-esm
* building tutorials in the documentation.
"""

from __future__ import annotations

DEFAULT_CATALOGS = {
    'aws_cesm2_le': 'https://raw.githubusercontent.com/intake/intake-esm/main/tutorial-catalogs/AWS-CESM2-LENS.json',
    'aws_cmip6': 'https://raw.githubusercontent.com/intake/intake-esm/main/tutorial-catalogs/AWS-CMIP6.json',
    'google_cmip6': 'https://raw.githubusercontent.com/intake/intake-esm/main/tutorial-catalogs/GOOGLE-CMIP6.json',
}


def get_url(name: str) -> str:
    """
    Get a small-catalog URL from the online repository (requires internet).
    If a local copy is found then always use that to avoid network traffic.
    Available catalogs:

    * ``"aws_cesm2_le"``
    * ``"aws_cmip6"``
    * ``"google_cmip6"''

    Parameters
    ----------
    name : str
        Name of the catalog. e.g. 'aws_cmip6'

    Returns
    -------
    str
        URL of the catalog.
    """

    if name in DEFAULT_CATALOGS:
        return DEFAULT_CATALOGS[name]

    raise KeyError(
        f'KeyError: {name} is an unknown key. Only small-catalogs in our `tutorial-catalogs`'
        f'directory are supported with this method. Valid values include: {get_available_cats()} .'
    )


def get_available_cats() -> list[str]:
    """
    Get a list of all supported small-catalog key names that map to URL from the online repository.

    Returns
    -------
    list[str]
        List of all supported small-catalog key names.
    """

    return list(DEFAULT_CATALOGS.keys())
